#include "sinqAxis.h"
#include "sinqController.h"
#include <math.h>
#include <unistd.h>

sinqAxis::sinqAxis(class sinqController *pC, int axis)
    : asynMotorAxis((asynMotorController *)pC, axis), pC_(pC) {

    initial_poll_ = true;
    watchdogMovActive_ = false;
    init_poll_counter_ = 0;
}

asynStatus sinqAxis::atFirstPoll() { return asynSuccess; }

asynStatus sinqAxis::poll(bool *moving) {
    // Local variable declaration
    asynStatus pl_status = asynSuccess;
    asynStatus poll_status = asynSuccess;

    // =========================================================================

    // If this poll is the initial poll, check if the parameter library has
    // already been initialized. If not, force EPCIS to repeat the poll until
    // the initialization is complete (or until a timeout is reached). Once the
    // parameter library has been initialized, read configuration data from the
    // motor controller into it.
    if (initial_poll_) {
        poll_status = atFirstPoll();
        if (poll_status == asynSuccess) {
            initial_poll_ = false;
        } else {
            // Send a message to the IOC shell every 10 trials.
            init_poll_counter_ += 1;
            if (init_poll_counter_ % 10 == 0) {
                asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                          "%s => line %d:\nRunning function 'atFirstPoll' "
                          "failed %d times with error %s.",
                          __PRETTY_FUNCTION__, __LINE__, init_poll_counter_,
                          pC_->stringifyAsynStatus(poll_status));
            }

            // Wait for 100 ms until trying the entire poll again
            usleep(100000);
            return poll_status;
        }
    }

    // The poll function is just a wrapper around doPoll and
    // handles mainly the callParamCallbacks() function. This wrapper is used
    // to make sure callParamCallbacks() is called in case of a premature
    // return.
    poll_status = doPoll(moving);

    // If the poll status is ok, reset the error indicators in the parameter
    // library
    if (poll_status == asynSuccess) {
        pl_status = setIntegerParam(pC_->motorStatusProblem_, false);
        if (pl_status != asynSuccess) {
            pC_->paramLibAccessFailed(pl_status, "motorStatusProblem_",
                                      __PRETTY_FUNCTION__, __LINE__);
        }
        pl_status = setIntegerParam(pC_->motorStatusCommsError_, false);
        if (pl_status != asynSuccess) {
            pC_->paramLibAccessFailed(pl_status, "motorStatusCommsError_",
                                      __PRETTY_FUNCTION__, __LINE__);
        }

        pl_status = setStringParam(pC_->motorMessageText_, "");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             __PRETTY_FUNCTION__, __LINE__);
        }
    }

    // According to the function documentation of asynMotorAxis::poll, this
    // function should be called at the end of a poll implementation.
    pl_status = callParamCallbacks();
    if (pl_status != asynSuccess) {
        // If we can't communicate with the parameter library, it doesn't make
        // sense to try and upstream this to the user -> Just log the error
        asynPrint(
            pC_->pasynUserSelf, ASYN_TRACE_ERROR,
            "%s => line %d:\ncallParamCallbacks failed with %s for axis %d",
            __PRETTY_FUNCTION__, __LINE__,
            pC_->stringifyAsynStatus(poll_status), axisNo_);
        poll_status = pl_status;
    }

    return poll_status;
}

asynStatus sinqAxis::doPoll(bool *moving) { return asynSuccess; }

asynStatus sinqAxis::move(double position, int relative, double minVelocity,
                          double maxVelocity, double acceleration) {

    double target_position = 0.0;

    // Status of parameter library operations
    asynStatus pl_status = asynSuccess;

    // =========================================================================

    // Calculate the (absolute) target position
    if (relative) {

        double motorPosition = 0.0;
        pl_status =
            pC_->getDoubleParam(axisNo_, pC_->motorPosition_, &motorPosition);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorPosition_",
                                             __PRETTY_FUNCTION__, __LINE__);
        }

        target_position = position + motorPosition;
    } else {
        target_position = position;
    }

    // Set the target position
    pl_status = setDoubleParam(pC_->motorTargetPosition_, target_position);
    if (pl_status != asynSuccess) {
        return pC_->paramLibAccessFailed(pl_status, "motorTargetPosition_",
                                         __PRETTY_FUNCTION__, __LINE__);
    }

    return doMove(position, relative, minVelocity, maxVelocity, acceleration);
}

asynStatus sinqAxis::doMove(double position, int relative, double minVelocity,
                            double maxVelocity, double acceleration) {
    return asynSuccess;
}

asynStatus sinqAxis::movementTimeoutWatchdog(bool moving) {
    asynStatus pl_status;

    // Not moving -> Watchdog inactive
    if (!moving) {
        watchdogMovActive_ = false;
        return asynSuccess;
    }

    if (!watchdogMovActive_) {

        // These parameters are only needed in this branch
        double motorPosition = 0.0;
        double motorTargetPosition = 0.0;
        double motorVelBase = 0.0;
        double motorAccel = 0.0;
        time_t timeContSpeed = 0;
        time_t timeAccel = 0;

        // Activate the watchdog
        watchdogMovActive_ = true;

        pl_status =
            pC_->getDoubleParam(axisNo_, pC_->motorPosition_, &motorPosition);
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorPosition",
                                             __PRETTY_FUNCTION__, __LINE__);
        }

        pl_status =
            pC_->getDoubleParam(axisNo_, pC_->motorVelBase_, &motorVelBase);
        // Only calculate timeContSpeed if the motorVelBase_ has been populated
        // with a sensible value (e.g. > 0)
        if (pl_status == asynSuccess && motorVelBase > 0.0) {

            pl_status = pC_->getDoubleParam(axisNo_, pC_->motorTargetPosition_,
                                            &motorTargetPosition);
            if (pl_status == asynSuccess) {
                timeContSpeed =
                    std::ceil(std::fabs(motorTargetPosition - motorPosition) /
                              motorVelBase);
            }
        }

        pl_status = pC_->getDoubleParam(axisNo_, pC_->motorAccel_, &motorAccel);
        if (pl_status == asynSuccess && motorVelBase > 0.0 &&
            motorAccel > 0.0) {
            timeAccel = 2 * std::ceil(motorVelBase / motorAccel);
        }

        // Calculate the expected arrival time
        expectedArrivalTime_ =
            time(NULL) + offsetMovTimeout_ +
            scaleMovTimeout_ * (timeContSpeed + 2 * timeAccel);

    } else if (expectedArrivalTime_ < time(NULL)) {
        // Check the watchdog
        asynPrint(pC_->pasynUserSelf, ASYN_TRACE_ERROR,
                  "%s => line %d:\nAxis %d exceeded the expected arrival time "
                  "(%ld).\n",
                  __PRETTY_FUNCTION__, __LINE__, axisNo_, expectedArrivalTime_);

        pl_status = setStringParam(
            pC_->motorMessageText_,
            "Exceeded expected arrival time. Check if the axis is blocked.");
        if (pl_status != asynSuccess) {
            return pC_->paramLibAccessFailed(pl_status, "motorMessageText_",
                                             __PRETTY_FUNCTION__, __LINE__);
        }

        pl_status = setIntegerParam(pC_->motorStatusProblem_, true);
        if (pl_status != asynSuccess) {
            pC_->paramLibAccessFailed(pl_status, "motorStatusProblem_",
                                      __PRETTY_FUNCTION__, __LINE__);
        }

        return asynError;
    }
    return asynSuccess;
}