/*
This class extends asynMotorAxis by some features used in SINQ.

Stefan Mathis, November 2024
*/
#ifndef __SINQDRIVER
#define __SINQDRIVER
#include "asynMotorAxis.h"

class epicsShareClass sinqAxis : public asynMotorAxis {
  public:
    sinqAxis(class sinqController *pC_, int axis);

    /**
    This function is executed at the very first poll after the IOC startup. If
    it returns anything else than 'asynSuccess', the function is evaluated again
    after 100 ms until it succeeds. Every 10 trials a warning is emitted.
     */
    asynStatus atFirstPoll();

    /**
    Wrapper around doPoll which performs the following operations;
    Before calling doPoll:
    - Try to execute atFirstPoll once (and retry, if that failed)

    After calling doPoll:
    - Reset motorStatusProblem_, motorStatusCommsError_ and motorMessageText_ if
    doPoll returned asynSuccess
    - Run `callParamCallbacks`
    - Return the status of `doPoll`
     */
    asynStatus poll(bool *moving);

    /**
    Implementation of the "proper", device-specific poll method. This method
    should be implemented by a child class of sinqAxis.
     */
    asynStatus doPoll(bool *moving);

    /**
    Wrapper around move which calculates the (absolute) target position and
    stores it in the parameter library. After that, it calls and returns doMove
     */
    asynStatus move(double position, int relative, double minVelocity,
                    double maxVelocity, double acceleration);

    /**
    Implementation of the "proper", device-specific poll method. This method
    should be implemented by a child class of sinqAxis.
    */
    asynStatus doMove(double position, int relative, double minVelocity,
                      double maxVelocity, double acceleration);

    /**
    Manages a timeout mechanism for the movement:
    If the movement takes too long, create an error message and return
    asynError. The watchdog is started when moving switches from "false" to
    "true" and stopped when moving switches from "true" to "false". At the
    watchdog start, the estimated movement time is calculated as

    t = offsetMovTimeout_ + scaleMovTime_ * [timeContSpeed + 2*timeAccel]

    with

    timeContSpeed = abs(motorTargetPosition - motorPosition) / motorVelBase
    timeAcc = motorVelBase / motorAccel

    The values motorTargetPosition, motorVelBase, motorAccel and
    positionAtMovementStart are taken from the parameter library. Therefore it
    is necessary to populate them before using this function. If they are not
    given, both speed and velocity are assumed to be infinity. This means that
    timeContSpeed and/or timeAcc are set to zero. motorTargetPosition is
    populated automatically when using the doMove function.

    The values offsetMovTimeout_ and scaleMovTimeout_ can be set directly from
    the IOC shell with the functions setScaleMovTimeout and setOffsetMovTimeout,
    if sinqMotor is loaded via the "require" mechanism.
     */
    asynStatus movementTimeoutWatchdog(bool moving);

    // Setter for offsetMovTimeout
    asynStatus setOffsetMovTimeout(time_t offsetMovTimeout) {
        offsetMovTimeout_ = offsetMovTimeout;
        return asynSuccess;
    }

    // Setter for scaleMovTimeout
    asynStatus setScaleMovTimeout(time_t scaleMovTimeout) {
        scaleMovTimeout_ = scaleMovTimeout;
        return asynSuccess;
    }

    friend class sinqController;

  protected:
    bool initial_poll_;
    int init_poll_counter_;

    // Helper variables for movementTimeoutWatchdog
    time_t expectedArrivalTime_;
    time_t offsetMovTimeout_;
    double scaleMovTimeout_;
    bool watchdogMovActive_;

  private:
    sinqController *pC_;
};

#endif
